<?php

/**
 * ClearOS framework configuration.
 *
 * @category   Framework
 * @package    Shared
 * @subpackage Libraries
 * @author     ClearFoundation <developer@clearfoundation.com>
 * @copyright  2011 ClearFoundation
 * @license    http://www.gnu.org/copyleft/lgpl.html GNU Lesser General Public License version 3 or later
 * @link       http://www.clearfoundation.com/docs/developer/framework/
 */

///////////////////////////////////////////////////////////////////////////////
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// N A M E S P A C E
///////////////////////////////////////////////////////////////////////////////

namespace clearos\framework;

///////////////////////////////////////////////////////////////////////////////
// D E P E N D E N C I E S
///////////////////////////////////////////////////////////////////////////////

use \clearos\framework\Logger as Logger;

require_once 'Logger.php';

///////////////////////////////////////////////////////////////////////////////
// C L A S S
///////////////////////////////////////////////////////////////////////////////

/**
 * ClearOS framework configuration.
 *
 * @category   Framework
 * @package    Shared
 * @subpackage Libraries
 * @author     ClearFoundation <developer@clearfoundation.com>
 * @copyright  2011 ClearFoundation
 * @license    http://www.gnu.org/copyleft/lgpl.html GNU Lesser General Public License version 3 or later
 * @link       http://www.clearfoundation.com/docs/developer/framework/
 */

class Config
{
    //-----------------------------------------------------------------------
    // V A R I A B L E S
    //-----------------------------------------------------------------------

    /**
     * @var array apps paths
     */

    public static $apps_paths = array('/usr/clearos/apps');

    /**
     * @var string base path for the framework
     */

    public static $framework_path = '/usr/clearos/framework';

    /**
     * @var string base path for the web server document root
     */

    public static $htdocs_path = '/usr/clearos/framework/htdocs';

    /**
     * @var string base path for themes
     */

    public static $themes_path = '/usr/clearos/themes';

    /**
     * @var boolean debug mode flag
     */

    public static $debug_mode = FALSE;

    /**
     * @var string debug log path
     */

    public static $debug_log = '/var/log/webconfig/framework_log';

    /**
     * @var array version handler for app developers
     */

    public static $clearos_devel_versions = array();

    /**
     * @var array scm path handler
     */

    public static $scm_subpaths = array('trunk', '');

    ///////////////////////////////////////////////////////////////////////////////
    // M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Config constructor.
     */

    public function __construct()
    {
        Logger::profile(__METHOD__, __LINE__);
    }

    /**
     * Returns the app base path.
     *
     * @param string $app app name
     *
     * @return string app base path
     */

    public static function get_app_base($app)
    {
        // FIXME: verbose
        // Logger::profile(__METHOD__, __LINE__);

        $version_paths = array('trunk', '');

        foreach (Config::$apps_paths as $path) {
            foreach ($version_paths as $version_path) {
                $base = $path . '/' . $app . '/' . $version_path;

                if (is_dir("$base/deploy"))
                    return $base;
            }
        }
    }

    /**
     * Returns the app base path.
     *
     * @param string $app app name
     *
     * @return string app base path
     */

    public static function get_app_root($app)
    {
        // Logging is verbose, don't bother
        $base_path = Config::get_app_base($app);

        if (preg_match('/.usr.clearos.apps/', $base_path)) {
            return '/approot';
        } else {
            $base_path = preg_replace('/\/webconfig\/apps\/.*/', '', $base_path);
            $base_path = preg_replace('/\/apps\/.*/', '', $base_path);
            $base_path = preg_replace('/.*\//', '', $base_path);

            return '/' . $base_path . '/approot';
        }
    }

    /**
     * Returns the app URL.
     *
     * @param string $app app name
     *
     * @return string app URL
     */

    public static function get_app_url($app)
    {
        Logger::profile(__METHOD__, __LINE__);

        $version_paths = array('trunk', '');

        $approot = Config::get_app_root($app);

        foreach (Config::$apps_paths as $path) {
            foreach ($version_paths as $version_path) {
                $base = $path . '/' . $app . '/' . $version_path;

                if (is_dir("$base/deploy"))
                    return $approot . '/' . $app . '/' . $version_path . '/htdocs';
            }
        }
    }

    /**
     * Returns the report driver name.
     *
     * @return string driver name
     */

    public static function get_reports_driver()
    {
        Logger::profile(__METHOD__, __LINE__);

        if (clearos_app_installed('professional_reports'))
            return 'professional_reports';
        else
            return 'home_reports';
    }

    /**
     * Returns the summary driver name.
     *
     * @return string driver name
     */

    public static function get_summary_driver()
    {
        Logger::profile(__METHOD__, __LINE__);

        // TODO: add support for other drivers
        return 'marketplace';
    }

    /**
     * Returns the theme path.
     *
     * @param string $theme theme name
     *
     * @return string theme path
     */

    public static function get_theme_path($theme)
    {
        Logger::profile(__METHOD__, __LINE__);

        if (isset(Config::$clearos_devel_versions['theme'][$theme]))
            $theme_version = '/' . Config::$clearos_devel_versions['theme'][$theme];
        else if (isset(Config::$clearos_devel_versions['theme']['default']))
            $theme_version = '/' . Config::$clearos_devel_versions['theme']['default'];
        else
            $theme_version = "";

        return Config::$themes_path . '/' . $theme . $theme_version;
    }

    /**
     * Returns the theme URL.
     *
     * @param string $theme theme name
     *
     * @return string theme URL
     */

    public static function get_theme_url($theme)
    {
        Logger::profile(__METHOD__, __LINE__);

        if (isset(Config::$clearos_devel_versions['theme'][$theme]))
            $theme_version = '/' . Config::$clearos_devel_versions['theme'][$theme];
        else if (isset(Config::$clearos_devel_versions['theme']['default']))
            $theme_version = '/' . Config::$clearos_devel_versions['theme']['default'];
        else
            $theme_version = "";

        return "/themes/" . $theme . $theme_version;
    }
}
